/* Copyright (c) 1996 University of Utah.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee is
 * hereby granted provided that this copyright notice appears in all
 * copies.
 *
 * See http://www.cs.utah.edu/~hamlet for more educational resources.
 *
 * The University of Utah MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The University of Utah SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

//*************************************************************************
//
//  Slider.java
//
//*************************************************************************

package SlidingBlock;

import java.awt.*;
import java.lang.Math;
import java.lang.ArithmeticException;
import hamlet.*;


public class Slider extends Canvas {

    // Colors

    final static Color boardColor = ColorList.BROWN;
    final static Color posbarColor = Color.red;
    Color blockColor;


    // Coordinates and lengths

    final static int blockSize = 7;
    final static int indicatorHalfLength = 5;
    final static int indicatorOffset = 3;
    final static int boardThickness = 3;

    final static int boardTopX = 2*indicatorHalfLength +
                                         2*indicatorOffset +
                                         boardThickness + 1;

    final static int boardTopY = blockSize*3/2;

    int boardLength;        // Length of board in pixels

    int yPosbar;            // Position of the vertical position bar
    int xPosbar;            // Position of the horizontal position bar

    double theta;           // Angle of ramp
    double cosTheta;
    double sinTheta;
    double factor;          // Net acceleration

  
    int deltaX;             // Horizontal and vertical offsets of top back
    int deltaY;             // corner of block from bottom back

    double delX;
    double delY;

    double mu;              // Coefficient of friction
    double time;            // Current simulated time
    double timeLimit;       // Time at which block reaches bottom of board

    double metersPerPixel;  // Number of meters per pixel

    double xPos;            // Horizonal and vertical coordinates of bottom
    double yPos;            // corner of trailing edge of block

    int xPosition;          // The values of xPos and yPos rounded to the
    int yPosition;          // nearest integer

    int xBlockPosition;     // We fudge the position of the block slightly
    int yBlockPosition;     // so that the background won't appear between
                            // the block and the ramp

    double trueDistance;    // Distance in meters block has moved

    int boardRun;           // Run and rise of the board in pixels
    int boardRise;

    ScrollableNumbers friction;    // Controls coefficient of friction
    NumericLabel distance;         // Displays distance moved by block

    NumLabel length;        // Board length displayed in meters
    NumLabel angle;         // Board angle displayed in degrees

    Polygon theBlock;       // Polygon representing the block.  Its
                            // lower-left corner is at (0,0)


    //--------------------------------------------------
    // Constructor
    //--------------------------------------------------

    Slider () {

	distance = null;
	friction = null;
	length = new NumLabel(this, 0, "m", 0, 0, Label.RIGHT);
	angle = new NumLabel(this, 0, "deg", 0, 0, Label.RIGHT);

	time = 0;
	mu = 0;
	metersPerPixel = 1;

	boardLength = -1;
	theta = -1;

    }


    //--------------------------------------------------
    // Make a copy of this slider
    //--------------------------------------------------

    Slider (Slider s) {

	distance = null;
	friction = null;
	length = new NumLabel(this, 0, "m", 0, 0, Label.RIGHT);
	angle = new NumLabel(this, 0, "deg", 0, 0, Label.RIGHT);

	synchronized (s) {
	    time = s.time;
	    mu = s.mu;
	    metersPerPixel = s.metersPerPixel;
	    setBoard(s.theta, s.boardLength);
	}

    }


    //--------------------------------------------------
    // Keep track of the distance display and friction control
    //--------------------------------------------------    

    public synchronized void setControls (ScrollableNumbers f, NumericLabel d)
    {
	distance = d;
	friction = f;
    }


    //--------------------------------------------------
    // Set up the prefered size
    //--------------------------------------------------

    public Dimension preferredSize () {
      return(new Dimension(270,230));
    }


    //--------------------------------------------------
    // Resets the angle and length of the board
    //--------------------------------------------------

    private void setBoard (double newTheta, int newLength) {

	// Make sure the new values are in bounds

	theta = newTheta;
	if (theta < 0) theta = 0;
	if (theta > Math.PI/2) theta = Math.PI/2;
	boardLength = newLength;

	// Calculate values directly dependent on theta

	cosTheta = Math.cos(theta);
	sinTheta = Math.sin(theta);
	delX = blockSize*sinTheta;
	delY = blockSize*cosTheta;
	deltaX = (int) Math.round(delX);
	deltaY = (int) Math.round(delY);

	// Create the block polygon

	theBlock = new Polygon();
	theBlock.addPoint(0, 0);
	theBlock.addPoint(deltaX, -deltaY);
	theBlock.addPoint(deltaX+deltaY, -deltaY+deltaX);
	theBlock.addPoint(deltaY, deltaX);

	// Calculate run and rise of the board

	boardRun = (int) Math.round(boardLength * cosTheta);
	boardRise = (int) Math.round(boardLength * sinTheta);

	// Determine location of position bars

	yPosbar = indicatorHalfLength + indicatorOffset;
	xPosbar = boardTopY + boardRise + deltaX + blockSize + 
	            indicatorHalfLength + indicatorOffset;

	// Set the block positions

	setPositions();

    }  


    //--------------------------------------------------
    // Set the distance scale
    //--------------------------------------------------

    public synchronized void setScale (double scale) {
	metersPerPixel = scale;
	setPositions();
	update(getGraphics());
    }


    private void setTime (double newTime) {
	time = newTime;
	computeDistance();
    }  


    private void setFriction(double newmu) {
	mu = newmu;
	setPositions();
    }
  


    //--------------------------------------------------
    // Draws the vertical and horizontal position bar indicators using
    // whatever color is current in g.  This makes it possible to do
    // erasures with these routines also.
    //--------------------------------------------------

    private void drawVertIndicator (int position, Graphics g) {
	g.drawLine(yPosbar-5, position, yPosbar+5, position);
    }


    private void eraseVertIndicator (int position, Graphics g) {
	if (position != boardTopY) {
	    g.setColor(getBackground());
	    drawVertIndicator(position, g);
	    g.setColor(posbarColor);
	    g.drawLine(yPosbar, position, yPosbar, position);
	}
    }


    private void drawHorizIndicator (int position, Graphics g) {
	g.drawLine(position, xPosbar-5, position, xPosbar+5);
    }


    private void eraseHorizIndicator (int position, Graphics g) {
	if (position != boardTopX) {
	    g.setColor(getBackground());
	    drawHorizIndicator(position, g);
	    g.setColor(posbarColor);
	    g.drawLine(position, xPosbar, position, xPosbar);
	}
    }
    

    //--------------------------------------------------
    // Draws the board.
    //--------------------------------------------------

    private void drawBoard (Graphics g) {

	g.setColor(boardColor);

	int run = boardRun + deltaY;
	int rise = boardRise + deltaX;

	Polygon board = new Polygon();
	board.addPoint(boardTopX, boardTopY);
	board.addPoint(boardTopX + run, boardTopY + rise);
	board.addPoint(boardTopX + run
		                 - (int) Math.round(boardThickness*sinTheta),
		       boardTopY + rise
		                 + (int) Math.round(boardThickness*cosTheta));
	board.addPoint(boardTopX - (int) Math.round(boardThickness*sinTheta),
		       boardTopY + (int) Math.round(boardThickness*cosTheta));

	g.fillPolygon(board);

    }
		       

    //--------------------------------------------------
    // Draws the box at the current position.
    //--------------------------------------------------

    private void drawBox (Graphics g) {
	g.setColor(blockColor);
	g.translate(xBlockPosition, yBlockPosition);
	g.fillPolygon(theBlock);
	g.translate(-xBlockPosition, -yBlockPosition);
    }


    //--------------------------------------------------
    // Updates the current animation.  This will only
    // work if time runs forward during animations.
    //--------------------------------------------------

    public synchronized boolean animationStep (double t) {

	if (t == time) return(false);

	if (t > timeLimit) {
	    if (time >= timeLimit) return(false);
	    else time = timeLimit;
	}

	// Record the old coordinates of the trailing corner of the box.

	int oldXPosition = xPosition;
	int oldYPosition = yPosition;
	int oldXBlockPosition = xBlockPosition;
	int oldYBlockPosition = yBlockPosition;
	

	// Update the time

	setTime(t);


	// If nothing has changed, do nothing further.  We are comparing
	// the integer versions of the positions here.  In we do
	// anti-aliasing in the future, we would want to compare the
	// double versions of the positions.

	if (timeLimit != time && xPosition == oldXPosition
	                      && yPosition == oldYPosition)
	    return(false);

	Graphics g = getGraphics();
	

	// Update the distance readout

	distance.setNumber(Math.round(trueDistance*100)/100.0);


	// Update the horizontal indicator if necessary.

	if (oldXPosition != xPosition) {
	    eraseHorizIndicator(oldXPosition, g);
	    drawHorizIndicator(xPosition, g);
	}


	// Update the vertical indicator if necessary.

	if (oldYPosition != yPosition) {
	    eraseVertIndicator(oldYPosition, g);
	    drawVertIndicator(yPosition, g);
	}


	// Obtain a rectangle that bounds the old and new positions
	// of the box.

	int left = oldXBlockPosition;
	int top = oldYBlockPosition - deltaY;
	int width = xBlockPosition + deltaX + deltaY;
	int height = yBlockPosition + deltaX - top;


	// Use the rectangle to create and splice in an updated image

	Image image = createImage(width+1, height+1);
	Graphics imageGraphics = image.getGraphics();
	imageGraphics.setColor(getBackground());
	imageGraphics.translate(-left, -top);
	imageGraphics.fillRect(left, top, width, height);
	drawBox(imageGraphics);
	drawBoard(imageGraphics);

	g.clipRect(left, top, width, height);
	g.drawImage(image, left, top, this);


	// Report that the block moved

	return(true);

    }



    //--------------------------------------------------
    // Does a single time step
    //--------------------------------------------------

    public synchronized void singleStep (double t) {
	if (t > timeLimit) t = timeLimit;
	setTime(t);
	update(getGraphics());
    }



    //--------------------------------------------------
    // Returns the current coefficient of friction
    //--------------------------------------------------

    public synchronized double friction () {
	return(mu);
    }

  
    //--------------------------------------------------
    // Get the maximum length of ramp that will fit in the winow
    //--------------------------------------------------


    public synchronized int getMaxLength (int height, int width) {

	// Determine the maximum rise and run

	int maxRise = height - 2*indicatorHalfLength - 2*blockSize
	                     - 2*indicatorOffset - 1 - boardTopY;

	int maxRun = width - 2*blockSize - boardTopX - deltaY;


	// Determine the maximum length

	double maxLength = Math.min(maxRise/Math.sin(theta),
				    maxRun/Math.cos(theta));

	return((int) Math.floor(maxLength));
	
    }



    //--------------------------------------------------
    // Create the graphics from scratch.  They are saved as an
    // offscreen image.
    //--------------------------------------------------

    public synchronized void paint (Graphics g) {

	// Draw a black line across the top

	g.setColor(Color.black);
	g.drawLine(0, 0, size().width, 0);


        // Draw position bars

	g.setColor(posbarColor);
	g.drawLine(yPosbar, boardTopY,
		   yPosbar, boardTopY+boardRise); 
	drawVertIndicator(boardTopY, g);
	drawVertIndicator(boardTopY+boardRise, g);

	g.drawLine(boardTopX, xPosbar,
		   boardTopX+boardRun, xPosbar);
	drawHorizIndicator(boardTopX, g);
	drawHorizIndicator(boardTopX+boardRun, g);


	// Draw current position indicator

	drawHorizIndicator(xPosition, g);
	drawVertIndicator(yPosition, g);


        // Draw the board and box.  It is important to do it in this
	// order because the box coordinates are fudged slightly to
	// overlap with the board.

	drawBox(g);
	drawBoard(g);


	// Update the distance readout

	distance.setNumber(Math.round(trueDistance*100)/100.0);


	// Draw the labels

	length.move(size().width-5, 2);
	length.setNumber(boardLength * metersPerPixel);
	length.paint(g);

	angle.move(size().width-5, length.height() + 4);
	angle.setNumber(Math.round(theta*180/Math.PI*100)/100.0);
	angle.paint(g);

    }


    //--------------------------------------------------
    // Reset to top of ramp
    //--------------------------------------------------

    public synchronized void reset () {
	setTime(0);
	update(getGraphics());
    }



    //--------------------------------------------------
    // Handles mouseDown events
    //--------------------------------------------------
    
    public synchronized boolean mouseDown (Event evt, int x, int y) {
	return(mouseDrag(evt, x, y));
    }

    
    
    //--------------------------------------------------
    // Handles mouseDrag events
    //--------------------------------------------------

    public synchronized boolean mouseDrag (Event evt, int x, int y) {

	if (x < boardTopX) x = boardTopX;

	else if (x > size().width - 2*blockSize)
	    x = size().width - 2*blockSize;

	if (y < boardTopY) y = boardTopY;

	else if (y > size().height - 2*indicatorHalfLength - 2*blockSize -
	                        2*indicatorOffset - 1)
	    y =  size().height - 2*indicatorHalfLength - 2*blockSize -
	                        2*indicatorOffset - 1;

	x = x - boardTopX;
        y = y - boardTopY;

	if (x == 0 && y == 0) return(true);
	
	setBoard(Math.atan2(y,x), (int) Math.sqrt(x*x + y*y));

	repaint();
	return true;
    }

    
    
    //--------------------------------------------------
    // Handles mouseUp events
    //--------------------------------------------------

    public synchronized boolean mouseUp (Event evt, int x, int y) {
	update(getGraphics());
	return true;
    }


    //--------------------------------------------------
    // Deals with the friction control
    //--------------------------------------------------

    public synchronized boolean action (Event theEvent, Object arg) {

	  if (theEvent.target instanceof ScrollableNumbers) {
	      setFriction(friction.getDoubleText());
	      update(getGraphics());
	      return true;
	  }

	return(false);

    }


    //--------------------------------------------------
    // Smoothly updates
    //--------------------------------------------------

    public synchronized void update (Graphics g) {
	Image offImage = createImage(size().width, size().height);
	Graphics g1 = offImage.getGraphics();
	g1.setColor(getBackground());
	g1.fillRect(0,0, bounds().width, bounds().height);
	g1.setFont(getFont());
	paint(g1);
	g.drawImage(offImage, 0, 0, this);
    }
      

    //--------------------------------------------------
    // Returns current time limit
    //--------------------------------------------------

    public synchronized double limit () {
	return(timeLimit);
    }
  


    //--------------------------------------------------
    // Sets the variables xPos and yPos, which are the horizontal
    // and vertical positions, respectively, of the lower-left corner
    // of the block.  The current time, ramp length, and ramp angle
    // are all taken into account.
    //--------------------------------------------------

    private void setPositions () {

	// Block color depends upon friction

	blockColor = new Color((int) (mu * 51),0,255 - (int) (mu * 51));


	// The time at which the block will reach the bottom
	// of the ramp.  Make sure the block doesn't move
	// backwards!

	factor = sinTheta - mu*cosTheta;
	if (factor < 0) factor = 0;
	
	if (factor == 0)
	    timeLimit = Double.MAX_VALUE;
	else
	    timeLimit = Math.sqrt(boardLength*metersPerPixel /
				  (4.9 * factor));


	computeDistance();

    }



    private void computeDistance () {

	// The distance the block has gone down the ramp

	trueDistance = 4.9 * time * time * factor;
	if (trueDistance < 0) trueDistance = 0;

	double dist = trueDistance / metersPerPixel;


	// The current horizontal position

	xPos = boardTopX + dist * cosTheta;
	if (xPos > boardTopX + boardRun) xPos = boardTopX+boardRun;
	xPosition = (int) Math.round(xPos);

	// The current vertical position

	yPos = boardTopY + dist * sinTheta;
	if (yPos > boardTopY+boardRise) yPos = boardTopY+boardRise;
	yPosition = (int) Math.round(yPos);

	// Fudge slightly so that the background won't show between
	// the block and the board.

	xBlockPosition = (int) Math.floor(xPosition - sinTheta);
	yBlockPosition = (int) Math.ceil(yPosition + cosTheta);

    }


    //--------------------------------------------------
    // Handle resize and reshape requests
    //--------------------------------------------------

    public void resize (int w, int h) {
	adjustSize(h, w);
	super.resize(w, h);
    }


    public void resize (Dimension d) {
	adjustSize(d.height, d.width);
	super.resize(d);
    }

    public void reshape (int x, int y, int w, int h) {
	adjustSize(h, w);
	super.reshape(x,y,w,h);
    }


    public synchronized void adjustSize (int h, int w) {

	// Remember the current angle and length

	double angle = theta;
	int length = boardLength;


	// If no angle exists, make it 45 degrees

	if (angle < 0) angle = Math.PI/4;
	

	// Figure out what the longest permissible length is.  If the
	// current length is larger, reset length.

	int maxLength = getMaxLength(h, w);
	if (length < 0 || length > maxLength) length = maxLength;


	// Set the angle and length

	setBoard(angle, length);

    }


}
